"""Miscellaneous utility methods used in a few applications"""
import numpy as np
from itertools import islice, chain


def avg_stdev(lst):
    """
    calculates, the average, standard deviation, and relative standard deviation of a list of values

    :param lst: list of values
    :return: average, standard deviation, relative standard deviation expressed as a percent
    """
    if len(lst) == 0:
        return None, None
    avg = sum(lst) / len(lst)
    if len(lst) >= 3:
        stdev = np.sqrt(
            sum([(i - avg) ** 2 for i in lst]) / (len(lst) - 1)
        )
    else:
        stdev = None
    if avg == 0. or stdev is None:
        pstdev = None
    else:
        pstdev = stdev / avg * 100.
    return avg, stdev, pstdev


def find_nearest(lst, value, wiggle=None):
    """
    Finds the nearest value in a dictionary or list

    :param lst: sorted list or dictionary with keys that are values
    :param value: value to find
    :param wiggle: the wiggle room that the value needs to be within (the bounds are [value-wiggle, value+wiggle])
    :return: the nearest key in the dictionary to the value
    """
    if len(lst) == 0:  # if there are no values
        return None
    if type(lst) == dict:  # if handed a dictionary
        lst = sorted(lst.keys())
    val = lst[
        np.abs(  # array of absolute differences of each list value to the target
            [val - value for val in lst]
        ).argmin()  # index of the minimum value
    ]
    if wiggle is not None and abs(value - val) > wiggle:  # if it's outside the wiggle area
        return None
    return val


def pad_list(lst: list,
        length: int,
        start_index: int = 0,
        pad_value=0.,
        ) -> list:
    """
    Pads a list to the specified length with the provided value starting and the specified index.

    :param lst: list to pad
    :param length: target length of the return list
    :param start_index: the starting index where to insert the pad
    :param pad_value: value to pad the list with
    :return: padded list
    :rtype: list
    """
    if start_index is 0:
        return [pad_value] * (length - len(lst)) + lst
    else:
        return lst[0:start_index] + [pad_value] * (length - len(lst)) + lst[start_index:]


def front_pad(lst: list,
              length: int,
              pad_value=0.,
              ) -> list:
    """
    Front pads a list to the specified length with the provided value.

    :param lst: list to pad
    :param length: target length
    :param pad_value: value to pad the list with
    :return: padded list
    :rtype: list
    """
    return [pad_value] * (length - len(lst)) + lst


def chunk_iterable(iterable, chunk_size):
    """
    Iterates an iterable in chunks of specified size.
    https://stackoverflow.com/questions/24527006/split-a-generator-into-chunks-without-pre-walking-it

    :param iterable: iterable to iterate over
    :param chunk_size: list chunk size
    :return: iterable
    """
    iterator = iter(iterable)
    for first in iterator:
        yield chain([first], islice(iterator, chunk_size - 1))



"""json interaction methods"""
import json


def create_json_file(json_file_path: str):
    """

    :param str, json_file_path: path to a json file
    :return:
    """
    if json_file_path[-5:] != '.json':
        json_file_path += '.json'

    with open(json_file_path, 'w+') as file:
        json.dump({}, file)

    return json_file_path


def load_json_file(json_file_path):
    json_file = open(json_file_path, "r")  # Open the JSON file for reading
    data = json.load(json_file)  # Read the JSON into the buffer
    json_file.close()  # Close the JSON file
    return data


def update_json_file(json_file_path: str,
                     data_to_update_with: dict):
    """
    update a json file with data. the data must be in the form of a dictionary.

    if the key of the dictionary value to add into the json file doesn't exist in the json file, create a new key
    value pair based on the data passed through.

    if the key of the dictionary value to add into the json file does exist in the json file and the value is a
    dictionary itself, update the current value to include the new dictionary value.

    if the key of the dictionary value to add into the json file does exist in the json file and the value is a
    list itself, update the current value to include the new list value.


    if the key of the dictionary value to add into the json file does exist in the json file and the value is a single
    value that isn't a dictionary or list, overwrite the value


    :param json_file_path:
    :param data_to_update_with:
    :return:
    """
    data_to_update_with_key = list(data_to_update_with.keys())[0]
    data_to_update_with_value = list(data_to_update_with.values())[0]
    data_to_update_with_value_type = type(data_to_update_with_value)

    data = load_json_file(json_file_path=json_file_path)

    existing_value_for_the_data_to_update_with_same_key = data.get(data_to_update_with_key)

    if existing_value_for_the_data_to_update_with_same_key is None:
        # if there is no existing value for the key you want to add to the json with, then just add it
        data[data_to_update_with_key] = data_to_update_with_value
    elif data_to_update_with_key is not None and data_to_update_with_value_type is dict:
        # if the key exists in the json but the value is  of the type dictionary, then add the new dictionary into
        #  the existing dictionary value
        existing_value_for_the_data_to_update_with_same_key.update(data_to_update_with_value)
    elif data_to_update_with_key is not None and data_to_update_with_value_type is list:
        # if the key exists in the json but the value is  of the type list, then add the new list values into
        #  the existing list value
        for item_in_list in data_to_update_with_value:
            existing_value_for_the_data_to_update_with_same_key.append(item_in_list)
    else:
        # if the key exists in the json, overwrite the value
        data[data_to_update_with_key] = data_to_update_with_value

    # Save changes to JSON file
    json_file = open(json_file_path, "w+")
    json_file.write(json.dumps(data))
    json_file.close()

    return json_file_path





import hein_utilities
from setuptools import setup, find_packages

NAME = 'hein_utilities'
AUTHOR = 'Lars Yunker / Hein Group'

PACKAGES = find_packages()
# KEYWORDS = ', '.join([
# ])

with open('LICENSE') as f:
    lic = f.read()
    lic.replace('\n', ' ')

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=NAME,
    version=hein_utilities.__version__,
    description='Methods and Classes used in a variety of Hein projects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    url='https://gitlab.com/heingroup/hein_utilities',
    packages=PACKAGES,
    license=lic,
    python_requires='>=3.6',
    project_urls={
        'Documentation': 'https://hein-utilities.readthedocs.io/en/latest/',
        'Hein Group': 'https://groups.chem.ubc.ca/jheints1/',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Operating System :: Microsoft :: Windows',
        'Natural Language :: English'
    ],
    # keywords=KEYWORDS,
    install_requires=[
        'numpy',
        'slackclient'
    ],
)



"""utility class for parsing and manipulating datetime objects"""
from datetime import datetime
from datetime import timedelta
from typing import List, Union
import warnings


class datetimeManager:
    _unit_seconds = 'seconds'
    _unit_minutes = 'minutes'
    _unit_hours = 'hours'
    _time_units = {0: _unit_seconds,
                   1: _unit_minutes,
                   2: _unit_hours,
                   }

    _reverse_time_units = dict((v, k) for k, v in _time_units.items())
    _default_datetime_format = '%Y_%m_%d_%H_%M_%S'  # to add in milliseconds, add %f

    def __init__(self,
                 string_format: str = _default_datetime_format,
                 ):
        """
        utility class for parsing and manipulating datetime objects

        :param string_format: datetime string format to use for parsing
        """
        self._string_format = string_format

    @property
    def string_format(self) -> str:
        return self._string_format

    @string_format.setter
    def string_format(self,
                      value: str,
                      ):
        if isinstance(value, str) is False:
            raise TypeError('value must be of type string')
        self._string_format = value

    @string_format.deleter
    def string_format(self) -> None:
        self._string_format = None

    @classmethod
    def str_to_datetime(cls,
                        *string_values: str,
                        datetime_format: str = None,
                        ) -> List[datetime]:
        """
        Convert a list of string values, back into datetime objects with a specific format; in this case, the string has
        to have been a datetime object that was converted into a string with the datetime_format that is passed in this
        function.

        Main use has previously been when files where timestamped, and when the file names need to be converted back
        into datetime objects in order to do calculations

        :param string_values: one or more (a list) of strings that can be converted into datetime objects
        :param str, datetime_format: a string to represent the datetime format the string should be converted into; it
            should also have been the format that the strings already are in
        :return:
        """
        if datetime_format is None:
            datetime_format = cls._default_datetime_format
        return [datetime.strptime(value, datetime_format) for value in string_values]

    @classmethod
    def relative_datetime(cls,
                          datetime_objects: List[datetime],
                          units: Union[_unit_seconds, _unit_minutes, _unit_hours] = None,
                          rounding: int = None,
                          ) -> List[timedelta]:
        """
        Convert an array of datetime objects that are absolute times, and return an array where all the times in the
        array are relative to the first time in the array. The relativity can be in seconds, minutes, or hours.

        :param datetime_objects: a list of datetime objects
        :param units: one of _unit_seconds, _unit_minutes, or _unit_hours
        :param int, rounding: the number of decimal places to round to
        :return:
        """
        # takes a list of datetime objects, and makes all the values relative to the first object in the list
        if units not in list(cls._time_units.values()):
            raise Exception(f'units must be one of: {list(cls._time_units.values())}')

        if units is None:
            units = cls._unit_seconds

        # make an array of timedelta objects where each value is the difference between the actual time relative to
        # the first time point
        array_of_datetime_timedelta = [datetime_value - datetime_objects[0] for datetime_value in
                                       datetime_objects]

        # convert the relative timedeltas to floats, where the float number is the number of seconds since the first
        # time point
        array_of_relative_x_in_seconds = [array_of_datetime_timedelta[index].total_seconds() for index
                                          in range(len(array_of_datetime_timedelta))]

        if units == cls._unit_seconds:
            array_of_relative_datetime_objects = array_of_relative_x_in_seconds
        elif units == cls._unit_minutes:
            array_of_relative_x_in_minutes = [array_of_relative_x_in_seconds[index] / 60.0 for index in
                                              range(len(array_of_relative_x_in_seconds))]
            array_of_relative_datetime_objects = array_of_relative_x_in_minutes
        elif units == cls._unit_hours:
            array_of_relative_x_in_hours = [array_of_relative_x_in_seconds[index] / 3600.0 for index in
                                            range(len(array_of_relative_x_in_seconds))]
            array_of_relative_datetime_objects = array_of_relative_x_in_hours

        if rounding is not None:
            array_of_relative_datetime_objects = [round(datetime_obj, rounding) for datetime_obj in
                                                  array_of_relative_datetime_objects]

        return array_of_relative_datetime_objects

    def now_string(self) -> str:
        """
        Get the current time from datetime.now() formatted as as a string according to the string_format property
        :return:
        """
        return datetime.now().strftime(self.string_format)


def array_of_str_to_array_of_datetime_objects(list_of_string_values,
                                              datetime_format: str = '%Y_%m_%d_%H_%M_%S',
                                              ) -> List[datetime]:
    """
    Legacy passthrough for deprecated function. Use the datetimeManager class method instead.
    """
    warnings.warn(  # v3.7.1
        'This method has been deprecated, use the class method in the datetimeManager class; use that instead',
        DeprecationWarning,
        stacklevel=2,
    )


def array_of_datetime_objects_to_array_of_relative_datetime_objects(array_of_datetime_objects,
                                                                    units,
                                                                    ) -> List[timedelta]:
    """
    Legacy passthrough for deprecated function. Use the datetimeManager class method instead.
    """
    warnings.warn(  # v3.7.1
        'This method has been deprecated, use the class method in the datetimeManager class; use that instead',
        DeprecationWarning,
        stacklevel=2,
    )










"""Slack control manager"""
import logging
import slack
import re
from typing import Union, List, Callable, Pattern
from .bots import RTMSlackBot, WebClientOverride
from .parsing import is_event, is_bot_message
from ..control_manager import ControlManager, do_nothing, return_empty


# todo consider using children
logger = logging.getLogger(__name__)

# pattern for help retrieval
help_pattern = re.compile('help', re.IGNORECASE)


class RTMControlManager(RTMSlackBot, ControlManager):
    # default message parsing methods
    DEFAULT_PARSING_METHODS = [
        is_event,
        is_bot_message,
    ]

    help_pattern = help_pattern

    def __init__(self,
                 user_member_ids: Union[str, List[str]] = None,
                 token: str = None,
                 channel_name: str = None,
                 channel_id: str = None,
                 auto_reconnect: bool = True,
                 ping_interval: int = 10,
                 start_action: Callable = do_nothing,
                 stop_action: Callable = do_nothing,
                 resume_action: Callable = do_nothing,
                 pause_action: Callable = do_nothing,
                 status_query: Callable = return_empty,
                 help_query: Callable = return_empty,
                 pre_parsing_methods: List[Callable] = None,
                 ):
        """
        A control manager for executing steps contextually based on slack trigger messages.

        :param str, user_member_ids: Slack user ID(s) for the slack bot to be able to message user. find this by going to
            a user's profile, click the three dots (...) and there is the member ID. Example of a slack member ID: ABCDEF.
        :param str token: token to connect to slack client
        :param str channel_name: channel to message on. for example, #channelname
        :param str channel_id: channel id to listen for events on. If set, only events on this channel will be be
            caught and handled. for example: C1234ABCD. Find more information on finding this here:
            https://stackoverflow.com/questions/40940327/what-is-the-simplest-way-to-find-a-slack-team-id-and-a-channel-id
        :param auto_reconnect: enable/disable RTM auto-reconnect
        :param ping_interval: ping interval for RTM client
        :param start_action: callable action to execute on a start trigger
        :param stop_action: callable action to execute on a stop trigger
        :param resume_action: callable action to execute on a resume trigger
        :param pause_action: callable action to execute on a pause trigger
        :param status_query: callable which returns a string representing the status of the thing being controlled
        :param help_query: callable which returns a help string for the Slack interface
        :param pre_parsing_methods: methods to execute on an RTM payload prior to executing targetted methods.
            These methods must accept an RTM payload and return a bool (False will proceed with action execution,
            True on any parsing method will fail to trigger the action methods). See slack_integration.parsing for
            examples.
        """
        # initialize bot and manager
        RTMSlackBot.__init__(
            self,
            user_member_ids=user_member_ids,
            token=token,
            channel_name=channel_name,
            channel_id=channel_id,
            auto_reconnect=auto_reconnect,
            ping_interval=ping_interval,
        )
        ControlManager.__init__(
            self,
            start_action=start_action,
            stop_action=stop_action,
            resume_action=resume_action,
            pause_action=pause_action,
            status_query=status_query,
        )

        # register catch methods
        self.register_message_catch(self.start_catch)
        self.register_message_catch(self.pause_catch)
        self.register_message_catch(self.stop_catch)
        self.register_message_catch(self.resume_catch)
        self.register_message_catch(self.status_catch)
        self.register_message_catch(self.help_catch)
        # store pre-parsing methods
        if pre_parsing_methods is None:
            pre_parsing_methods = self.DEFAULT_PARSING_METHODS
        self.parsing_methods: List[Callable] = pre_parsing_methods

        self._help_query = return_empty
        self.help_query = help_query

    @property
    def help_query(self) -> Callable:
        """the callable which is used to retrieve a help string"""
        return self._help_query

    @help_query.setter
    def help_query(self, value):
        if callable(value) is False:
            raise TypeError(f'{value} is not callable')
        self._help_query = value

    @help_query.deleter
    def help_query(self):
        self._help_query = return_empty

    def catch_and_match(self, pattern: Union[Pattern, str], full_match: bool = True, **payload) -> bool:
        """
        Attempts to match a regex pattern to the message provided

        :param pattern: regex pattern to compare against the text, or a string that will be compiled into a regex
            pattern with the re.IGNORECASE flag added
        :param full_match:  if True, then the payload is text has to re.fullmatch the pattern to return True,
            else, as long as the pattern is found somewhere in the text (re.search), then the return value will be True
        :param payload: RTM payload
        :return: match truth
        """
        # apply pre-parsing methods
        if any([pre_parse(**payload) is True for pre_parse in self.parsing_methods]):
            return False
        message = payload['data']
        text = message.get('text')
        channel_id = str(message.get('channel'))
        if self.channel_id is not None:
            if self.channel_id != channel_id:
                return False
        if type(pattern) == str:
            pattern = re.compile(pattern, re.IGNORECASE)
        if full_match:
            if pattern.fullmatch(text) is not None:
                logger.debug(f'{pattern} catch triggered by "{text}"')
                return True
        else:
            if pattern.search(text) is not None:
                logger.debug(f'{pattern} catch triggered by "{text}"')
                return True
        return False

    def start_catch(self, **payload):
        """Catches start calls and executes the start action"""
        if self.catch_and_match(self.start_pattern, **payload) is True:
            with WebClientOverride(self, payload['web_client']):
                if self.start_action is not do_nothing:
                    self.post_slack_message(
                        'Sure thing! Starting things up! :science_parrot:'
                    )
                    # assumes non-blocking
                    self.start_action()

    def pause_catch(self, **payload):
        """catches pause calls and executes the pause action"""
        if self.catch_and_match(self.pause_pattern, **payload) is True:
            with WebClientOverride(self, payload['web_client']):
                if self.pause_action is not do_nothing:
                    self.post_slack_message(
                        'Pausing the thing!'
                    )
                    # assumes non-blocking
                    self.pause_action()

    def resume_catch(self, **payload):
        """catches resume calls and execute the resume action"""
        if self.catch_and_match(self.resume_pattern, **payload) is True:
            with WebClientOverride(self, payload['web_client']):
                if self.resume_action is not do_nothing:
                    self.post_slack_message(
                        'Resuming the thing!'
                    )
                    # assumes non-blocking
                    self.resume_action()

    def stop_catch(self, **payload):
        """catches stop calls and executes the stop action"""
        if self.catch_and_match(self.stop_pattern, **payload) is True:
            with WebClientOverride(self, payload['web_client']):
                if self.stop_action is not do_nothing:
                    self.post_slack_message(
                        'Stopping the thing!'
                    )
                    # assumes non-blocking
                    self.stop_action()

    def status_catch(self, **payload):
        """catches and posts the status query string"""
        if self.catch_and_match(self.status_pattern, **payload) is True:
            with WebClientOverride(self, payload['web_client']):
                if self.status_query is not return_empty:
                    self.post_slack_message(
                        f"Here's the current status:\n`{self.status_query()}`"
                    )

    def help_catch(self, **payload):
        """catches and posts the help string"""
        if self.catch_and_match(self.help_pattern, **payload) is True:
            with WebClientOverride(self, payload['web_client']):
                # todo
                #   - iterate through defined
                #   - check if defined
                #   - add to help string
                if self.help_query is not return_empty:
                    self.post_slack_message(
                        f"Here's how to use the Slack interface:\n`{self.help_query()}`"
                    )



"""
Contextual action managers
"""
import re
import logging
from typing import Callable, Pattern

# regex patterns
start_pattern = re.compile('start', re.IGNORECASE)
pause_pattern = re.compile('pause|suspend', re.IGNORECASE)
stop_pattern = re.compile('stop|halt|quit', re.IGNORECASE)
resume_pattern = re.compile('resume', re.IGNORECASE)
status_pattern = re.compile('status', re.IGNORECASE)

# logger
logger = logging.getLogger(__name__)


def do_nothing(*args, **kwargs):
    """function that does nothing"""
    logger.debug('doing nothing')
    pass


def return_empty(*args, **kwargs):
    """function that returns an empty string"""
    return ''


class ControlManager:
    # catch patterns
    start_pattern: Pattern = start_pattern
    pause_pattern: Pattern = pause_pattern
    stop_pattern: Pattern = stop_pattern
    resume_pattern: Pattern = resume_pattern
    status_pattern: Pattern = status_pattern

    def __init__(self,
                 start_action: Callable = do_nothing,
                 stop_action: Callable = do_nothing,
                 resume_action: Callable = do_nothing,
                 pause_action: Callable = do_nothing,
                 status_query: Callable = return_empty,
                 ):
        """
        A control manager for executing steps contextually. Context actions can be assigned as properties of the manager
        so that they can be executed by the appropriate trigger.

        :param start_action: callable action to execute on a start trigger
        :param stop_action: callable action to execute on a stop trigger
        :param resume_action: callable action to execute on a resume trigger
        :param pause_action: callable action to execute on a pause trigger
        :param status_query: callable which returns a string representing the status of the thing being controlled
        """
        self._start_action = do_nothing
        self._pause_action = do_nothing
        self._resume_action = do_nothing
        self._stop_action = do_nothing
        self._status_query = return_empty
        self.start_action = start_action
        self.stop_action = stop_action
        self.resume_action = resume_action
        self.pause_action = pause_action
        self.status_query = status_query

    def __str__(self):
        return f'{self.__class__.__name__}'

    @property
    def start_action(self) -> Callable:
        """the callable which is executed on start catches"""
        return self._start_action

    @start_action.setter
    def start_action(self, value: Callable):
        if callable(value) is False:
            raise TypeError(f'{value} is not callable')
        self._start_action = value

    @start_action.deleter
    def start_action(self):
        self._start_action = do_nothing

    @property
    def pause_action(self) -> Callable:
        """the callable which is executed on pause catches"""
        return self._pause_action

    @pause_action.setter
    def pause_action(self, value: Callable):
        if callable(value) is False:
            raise TypeError(f'{value} is not callable')
        self._pause_action = value

    @pause_action.deleter
    def pause_action(self):
        self._pause_action = do_nothing

    @property
    def resume_action(self) -> Callable:
        """the callable which is executed on resume catches"""
        return self._resume_action

    @resume_action.setter
    def resume_action(self, value: Callable):
        if callable(value) is False:
            raise TypeError(f'{value} is not callable')
        self._resume_action = value

    @resume_action.deleter
    def resume_action(self):
        self._resume_action = do_nothing

    @property
    def stop_action(self) -> Callable:
        """the callable which is executed on stop catches"""
        return self._stop_action

    @stop_action.setter
    def stop_action(self, value: Callable):
        if callable(value) is False:
            raise TypeError(f'{value} is not callable')
        self._stop_action = value

    @stop_action.deleter
    def stop_action(self):
        self._stop_action = do_nothing

    @property
    def status_query(self) -> Callable:
        """the callable which is used to retrieve a status string"""
        return self._status_query

    @status_query.setter
    def status_query(self, value: Callable):
        if callable(value) is False:
            raise TypeError(f'{value} is not callable')
        self._status_query = value

    @status_query.deleter
    def status_query(self):
        self._status_query = return_empty
